message tcp_stream_entry {
	required uint32		inq_len		= 1;
	required uint32		inq_seq		= 2;
	required uint32		outq_len	= 3;
	required uint32		outq_seq	= 4;

	required uint32		opt_mask	= 5; /* TCPI_OPT_ bits */
	required uint32		snd_wscale	= 6;
	required uint32		mss_clamp	= 7;
	optional uint32		rcv_wscale	= 8;
}
