message mnt_entry {
	required uint32		fstype			= 1;
	required uint32		mnt_id			= 2;
	required uint32		root_dev		= 3;
	required uint32		parent_mnt_id		= 4;
	required uint32		flags			= 5;

	required string		root			= 6;
	required string		mountpoint		= 7;
	required string		source			= 8;
	required string		options			= 9;
}
