#
# System calls table, please make sure the table consist only the syscalls
# really used somewhere in project.
#
# The template is (name and srguments are optinal if you need only __NR_x
# defined, but no realy entry point in syscalls lib).
#
# name			code		name			arguments
# -----------------------------------------------------------------------
#
__NR_read		0		sys_read		(int fd, void *buf, unsigned long count)
__NR_write		1		sys_write		(int fd, const void *buf, unsigned long count)
__NR_open		2		sys_open		(const char *filename, unsigned long flags, unsigned long mode)
__NR_close		3		sys_close		(int fd)
__NR_lseek		8		sys_lseek		(int fd, unsigned long offset, unsigned long origin)
__NR_mmap		9		sys_mmap		(void *addr, unsigned long len, unsigned long prot, unsigned long flags, unsigned long fd, unsigned long offset)
__NR_mprotect		10		sys_mprotect		(const void *addr, unsigned long len, unsigned long prot)
__NR_munmap		11		sys_munmap		(void *addr, unsigned long len)
__NR_brk		12		sys_brk			(void *addr)
__NR_rt_sigaction	13		sys_sigaction		(int signum, const rt_sigaction_t *act, rt_sigaction_t *oldact, size_t sigsetsize)
__NR_rt_sigprocmask	14		sys_sigprocmask		(int how, k_rtsigset_t *set, k_rtsigset_t *old, size_t sigsetsize)
__NR_rt_sigreturn	15		sys_rt_sigreturn	(void)
__NR_mincore		27		sys_mincore		(void *addr, unsigned long size, unsigned char *vec)
__NR_shmat		30		sys_shmat		(int shmid, void *shmaddr, int shmflag)
__NR_pause		34		sys_pause		(void)
__NR_nanosleep		35		sys_nanosleep		(struct timespec *req, struct timespec *rem)
__NR_getitimer		36		sys_getitimer		(int which, const struct itimerval *val)
__NR_setitimer		38		sys_setitimer		(int which, const struct itimerval *val, struct itimerval *old)
__NR_getpid		39		sys_getpid		(void)
__NR_socket		41		sys_socket		(int domain, int type, int protocol)
__NR_connect		42		sys_connect		(int sockfd, struct sockaddr *addr, int addrlen)
__NR_sendmsg		46		sys_sendmsg		(int sockfd, const struct msghdr *msg, int flags)
__NR_recvmsg		47		sys_recvmsg		(int sockfd, struct msghdr *msg, int flags)
__NR_bind		49		sys_bind		(int sockfd, const struct sockaddr *addr, int addrlen)
__NR_setsockopt		54		sys_setsockopt		(int sockfd, int level, int optname, const void *optval, socklen_t optlen)
__NR_getsockopt		55		sys_getsockopt		(int sockfd, int level, int optname, const void *optval, socklen_t *optlen)
__NR_clone		56		sys_clone		(unsigned long flags, void *child_stack, void *parent_tid, void *child_tid)
__NR_exit		60		sys_exit		(unsigned long error_code)
__NR_wait4		61		sys_waitpid		(int pid, int *status, int options, struct rusage *ru)
__NR_kill		62		sys_kill		(long pid, int sig)
__NR_fcntl		72		sys_fcntl		(int fd, int type, long arg)
__NR_flock		73		sys_flock		(int fd, unsigned long cmd)
__NR_mkdir		83		sys_mkdir		(const char *name, int mode)
__NR_rmdir		84		sys_rmdir		(const char *name)
__NR_unlink		87		sys_unlink		(char *pathname)
__NR_readlink		89		sys_readlink		(const char *path, char *buf, int bufsize)
__NR_setresuid		117		sys_setresuid		(int uid, int euid, int suid)
__NR_setresgid		119		sys_setresgid		(int gid, int egid, int sgid)
__NR_getpgid		121		sys_getpgid		(void)
__NR_setfsuid		122		sys_setfsuid		(int fsuid)
__NR_setfsgid		123		sys_setfsgid		(int fsgid)
__NR_getsid		124		sys_getsid		(void)
__NR_capset		126		sys_capset		(struct cap_header *h, struct cap_data *d)
__NR_personality	135		sys_personality		(unsigned int personality)
__NR_prctl		157		sys_prctl		(int option, unsigned long arg2, unsigned long arg3, unsigned long arg4, unsigned long arg5)
__NR_arch_prctl		158		sys_arch_prctl		(int option, unsigned long addr)
__NR_mount		165		sys_mount		(char *dev_nmae, char *dir_name, char *type, unsigned long flags, void *data)
__NR_umount2		166		sys_umount2		(char *name, int flags)
__NR_gettid		186		sys_gettid		(void)
__NR_futex		202		sys_futex		(u32 *uaddr, int op, u32 val, struct timespec *utime, u32 *uaddr2, u32 val3)
__NR_set_thread_area	205		sys_set_thread_area	(user_desc_t *info)
__NR_get_thread_area	211		sys_get_thread_area	(user_desc_t *info)
__NR_set_tid_address	218		sys_set_tid_address	(int *tid_addr)
__NR_restart_syscall	219		sys_restart_syscall	(void)
__NR_exit_group		231		sys_exit_group		(int error_code)
__NR_set_robust_list	273		sys_set_robust_list	(struct robust_list_head *head, size_t len)
__NR_get_robust_list	274		sys_get_robust_list	(int pid, struct robust_list_head **head_ptr, size_t *len_ptr)
__NR_open_by_handle_at	304		sys_open_by_handle_at	(int mountdirfd, struct file_handle *handle, int flags)
__NR_setns		308		sys_setns		(int fd, int nstype)
__NR_kcmp		312		sys_kcmp		(pid_t pid1, pid_t pid2, int type, unsigned long idx1, unsigned long idx2)
