import "fown.proto";

message eventpoll_tfd_entry {
	required uint32		id	= 1;
	required uint32		tfd	= 2;
	required uint32		events	= 3;
	required uint64		data	= 4;
}

message eventpoll_file_entry {
	required uint32		id	= 1;
	required uint32		flags	= 2;
	required fown_entry	fown	= 3;
}
